def pytest_configure():
    from django.conf import settings

    settings.configure(
        DEBUG_PROPAGATE_EXCEPTIONS=True,
        DEBUG = True,
        DATABASES={
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:'
            }
        },
        SITE_ID=1,
        SECRET_KEY='not very secret in tests',
        USE_I18N=True,
        USE_L10N=True,
        STATIC_URL='/static/',
        ROOT_URLCONF='tests.urls',
        TEMPLATES=[
            {
                'BACKEND': 'django.template.backends.django.DjangoTemplates',
                'APP_DIRS': True,
            },
        ],
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.middleware.common.CommonMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.contrib.messages.middleware.MessageMiddleware',
        ),
        INSTALLED_APPS=(
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.sites',
            'django.contrib.staticfiles',
            'rest_framework',
            'knox',
            'tests',
        ),
        PASSWORD_HASHERS=(
            'django.contrib.auth.hashers.MD5PasswordHasher',
        ),
        REST_FRAMEWORK={
            'DEFAULT_AUTHENTICATION_CLASSES': [
                'rest_framework.authentication.BasicAuthentication',
                'rest_framework.authentication.SessionAuthentication',
            ],
        },
    )

    try:
        import django
        django.setup()
    except AttributeError:
        pass



import random
from datetime import timedelta

from django.db.models import F
from knox.models import AuthToken
from rest_framework.test import APITestCase

from django.contrib.auth.models import User
from django.urls import reverse
from rest_framework import status

from jw_nx.tokens import AccessToken, RefreshToken

verify_url = reverse('jw-nx-verify')
login_url = reverse('jw-nx-login')
refresh_url = reverse('jw-nx-refresh')
logout_current_url = reverse('jw-nx-logout')
logout_other_url = reverse('jw-nx-logout-other')
logout_all_url = reverse('jw-nx-logout-all')
average_per_user_url = reverse('admin-average-all-per-user')
average_expired_per_user_url = reverse('admin-average-expired-per-user')
average_active_per_user_url = reverse('admin-average-active-per-user')
delete_expired_tokens_url = reverse('admin-delete-expired-tokens')


class APIAuthTest(APITestCase):
    user_count = 0
    password = "TestPasswordForAllUsers123"

    def create_test_user(self):
        """ Adds the default user to the database """
        user = User.objects.create_user(username=f'user{self.user_count}',
                                        email=f"user{self.user_count}@test.com",
                                        password=self.password)
        self.user_count += 1
        return user

    def create_test_admin(self):
        """ Adds the default admin to the database """
        admin = User.objects.create_superuser(username=f"admin{self.user_count}",
                                              email='admin@test.gmail', is_staff=True,
                                              password=self.password)
        self.user_count += 1
        self.assertTrue(admin.is_superuser)
        self.assertTrue(admin.is_staff)
        return admin

    def with_token(self, token):
        if not token:
            self.client.credentials(HTTP_AUTHORIZATION=None)
            return self

        self.client.credentials(HTTP_AUTHORIZATION="Bearer " + token)
        return self

    def login(self, user=None):
        user = user if user is not None else self.create_test_user()
        payload = {
            'username': user.username,
            'password': self.password
        }

        response = self.client.post(login_url, data=payload, format='json')
        self.assertIn('access_token', response.data)
        self.assertIn('refresh_token', response.data)
        access = response.data['access_token']
        refresh = response.data['refresh_token']

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone(access)
        self.assertIsNotNone(refresh)

        self.assertTrue(user.is_authenticated)
        return access, refresh

    def test_login(self):
        """ Test that login view is creating access_token and refresh_token """
        with self.assertNumQueries(3):
            """
             Expected queries:
             1- Create user
             2- Retrieve user by authentication credentials
             3- Create knox token in validate method of LoginSerializer
            """
            user = self.create_test_user()
            ac, re = self.login(user)

    def test_verify(self):
        """ Test 'verify' endpoint """
        user = self.create_test_user()
        ac, re = self.login(user)

        with self.assertNumQueries(2):
            """
             Expected queries:
             1- Retrieve user by authentication credentials
             2- Retrieve knox token in validate_jkt
            """
            response = self.with_token(ac).client.post(verify_url)

            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

    def test_verify_without_bearer(self):
        """ Test that authentication without bearer header is fail """
        response = self.client.post(verify_url)

        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response.data['detail'].title(), "Authentication Credentials Were Not Provided.")

    def test_verify_with_invalid_access_token(self):
        """ Test verify endpoint with invalid access token """
        user = self.create_test_user()
        ac, re = self.login(user)
        ac += '0'
        with self.assertNumQueries(0):
            response = self.with_token(ac).client.post(verify_url)

            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
            self.assertEqual(response.data['detail'].title().lower(), 'Error decoding signature.'.lower())

    def test_verify_with_wrong_jkt(self):
        """ Test verify endpoint with invalid knox_token(jkt claim) """
        user = self.create_test_user()
        ac, re = self.login(user)
        # Update jkt claim value
        access = AccessToken()
        access.payload = access.decode(ac)
        access.payload['jkt'] = 'Invalid1Knox2Token'

        with self.assertNumQueries(2):
            """
             Expected queries:
             1- Retrieve user by authentication credentials
             2- Retrieve knox by invalid jkt token_key
            """
            response = self.with_token(str(access)).client.post(verify_url)

            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
            self.assertEqual(response.data['detail'].title().lower(), 'Invalid Token'.lower())

    def test_verify_with_expired_token(self):
        """ Test that expired knox token is not valid """
        user = self.create_test_user()
        ac, re = self.login(user)
        AuthToken.objects.filter(user=user).update(expiry=F('expiry') - (F('expiry') + timedelta(days=1)))

        response = self.with_token(ac).client.post(verify_url)

        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response.data['detail'].title().lower(), 'Invalid Token'.lower())

    def test_verify_with_expired_access_token(self):
        """ Test that expired access token is not valid """
        user = self.create_test_user()
        ac, re = self.login(user)

        # Update expiration time
        access = AccessToken()
        access.payload = access.decode(ac)
        access.payload['exp'] -= 60 * 60 * 24  # One day ago

        with self.assertNumQueries(0):
            response = self.with_token(str(access)).client.post(verify_url)

        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response.data['detail'].title().lower(), 'Signature has expired.'.lower())

    def test_refresh(self):
        """ Test refresh endpoint with valid refresh_token that should return `new` and `valid` access_token """
        user = self.create_test_user()
        ac, re = self.login(user)
        payload = {"refresh_token": re}
        response = self.client.post(refresh_url, data=payload, format='json')

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('access_token', response.data)
        self.assertIn('refresh_token', response.data)

        check_response = self.with_token(response.data['access_token']).client.get(verify_url)
        self.assertEqual(check_response.status_code, status.HTTP_204_NO_CONTENT)

    def test_refresh_invalid(self):
        """ Test that refresh endpoint with invalid refresh token is not return access token """
        user = self.create_test_user()
        ac, re = self.login(user)
        re = f"{re}invalid"

        with self.assertNumQueries(0):
            payload = {'refresh_token': re}
            response = self.client.post(refresh_url, data=payload, format='json')

        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_refresh_invalid_user_id(self):
        """ Test that refresh token with invalid user_id claim is not valid """
        ac, re = self.login()
        refresh = RefreshToken()
        refresh.validate_token(re)
        refresh.payload['user_id'] = 123

        with self.assertNumQueries(1):
            """
             Expected queries:
             1- Get knox token based on 'jkt', 'user_id' claim
            """
            payload = {'refresh_token': str(refresh)}
            response = self.client.post(refresh_url, data=payload, format='json')

        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response.data['detail'].title(), 'Invalid Token')
        self.assertNotIn('access_token', response.data)

    def test_refresh_invalid_jtk(self):
        """ Test that refresh token with invalid `jtk` claim is not valid """
        ac, re = self.login()
        refresh = RefreshToken()
        refresh.validate_token(re)
        refresh.payload['jkt'] = 'Invalid1Token2Key3'

        with self.assertNumQueries(1):
            """
             Expected queries:
             1- Get knox token based on 'jkt', 'user_id' claim
            """
            payload = {'refresh_token': str(refresh)}
            response = self.client.post(refresh_url, data=payload, format='json')

        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response.data['detail'].title(), 'Invalid Token')
        self.assertNotIn('access_token', response.data)

    def test_logout_current(self):
        """ Test that logout endpoint is working correctly """
        user = self.create_test_user()
        ac, re = self.login(user)

        with self.assertNumQueries(5):
            """
             Expected queries:
             1- Retrieve user by authentication credentials
             2- Retrieve knox token by 'jtk' claim
             3- Delete knox token from database
             ----------
             4- Retrieve user by authentication credentials
             5- Retrieve knox token by 'jkt' claim
            """
            response = self.with_token(ac).client.post(logout_current_url)
            check_response = self.with_token(ac).client.post(verify_url)

            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
            self.assertEqual(check_response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_logout_other(self):
        """ Test that logout other endpoint is working correctly """
        user = self.create_test_user()
        ac1, re1 = self.login(user)
        ac2, re2 = self.login(user)
        ac3, re3 = self.login(user)

        with self.assertNumQueries(3):
            """
             Expected queries:
             1- Retrieve user by authentication credentials
             2- Retrieve knox token by 'jtk' claim
             3- Delete excluded knox tokens
            """
            response = self.with_token(ac1).client.post(logout_other_url)
        response_check1 = self.with_token(ac1).client.post(verify_url)
        response_check2 = self.with_token(ac2).client.post(verify_url)
        response_check3 = self.with_token(ac3).client.post(verify_url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response_check1.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(response_check2.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response_check3.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_logout_all(self):
        """ Test that logout all endpoint is working correctly """
        user = self.create_test_user()
        ac1, re1 = self.login(user)
        ac2, re2 = self.login(user)
        ac3, re3 = self.login(user)

        with self.assertNumQueries(3):
            """
             Expected queries:
             1- Retrieve user by authentication credentials
             2- Retrieve knox token by 'jtk' claim
             3- Delete excluded knox tokens
            """
            response = self.with_token(ac1).client.post(logout_all_url)
        response_check1 = self.with_token(ac1).client.post(verify_url)
        response_check2 = self.with_token(ac2).client.post(verify_url)
        response_check3 = self.with_token(ac3).client.post(verify_url)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response_check1.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response_check2.status_code, status.HTTP_401_UNAUTHORIZED)
        self.assertEqual(response_check3.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_delete_expired_tokens(self):
        """ Test that all expired tokens are deleting by this endpoint """
        user = self.create_test_user()
        self.login(user)
        self.login(user)
        self.login(user)
        admin = self.create_test_admin()
        ac, re = self.login(admin)

        # expire all tokens that belong to this user
        AuthToken.objects.filter(user_id=user.id).update(expiry=F('expiry') - (F('expiry') + timedelta(hours=10)))

        response = self.with_token(ac).client.post(delete_expired_tokens_url)

        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertEqual(response.data['Count'], 3)

    def test_average_per_user(self):
        """ Test average_per_user endpoint that is working correctly """
        users_token_count = {}
        expiry = timedelta(minutes=10)

        for i in range(0, random.randint(1, 20)):  # Create user
            user = self.create_test_user()
            users_token_count.update({user.username: 0})
            for j in range(0, random.randint(1, 50)):
                AuthToken.objects.create(user, expiry)
                users_token_count[user.username] += 1

        admin = self.create_test_admin()  # This line is create 1 user and 1 extra token
        ac, re = self.login(admin)
        with self.assertNumQueries(3):
            """
             Expected queries:
             1- Retrieve user by ID from 'user_id' claim
             2- Retrieve knox token by 'jtk' claim
             3- Get average of expired counts of tokens for every user
            """
            response = self.with_token(ac).client.get(average_per_user_url)

        # Calculating average
        filtered_vals = [value for name, value in users_token_count.items()]
        average = (sum(filtered_vals) + 1) / (len(filtered_vals) + 1)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['Average'], average)

    def test_average_expired_per_user(self):
        """ Test that 'average expired per user' endpoint is getting average count of expired tokens per user"""
        users_list = []  # ['username1', 'username2', .. ]
        users_tokens = {}  # {'username' :[id1, id2, id3, ...]},{},{}
        expiry = timedelta(minutes=10)
        users_count = random.randint(1, 20)
        for i in range(0, users_count):  # Create user
            user = self.create_test_user()
            users_list.append(user.username)
            users_tokens.update({user.username: []})
            pks = []
            for j in range(0, random.randint(1, 50)):
                instance, _ = AuthToken.objects.create(user, expiry)
                pks.append(instance.pk)
            users_tokens[user.username] = pks

        users_expired_token_pks = {}  # {'username1': ['pk1', 'pk2', 'pk3'], 'username2': [.....]}
        for username in users_list:  # Expire random token of users
            users_expired_token_pks.update({username: []})
            all_tokens_of_user = users_tokens[username]
            expired_tokens_of_user = users_expired_token_pks[username]
            random_list = random.sample(range(0, len(all_tokens_of_user)), random.randint(1, len(all_tokens_of_user)))
            for random_index in random_list:
                chosen_pk_of_user = all_tokens_of_user[random_index]
                expired_tokens_of_user.append(chosen_pk_of_user)
                # Expire token
                AuthToken.objects \
                    .filter(pk=chosen_pk_of_user) \
                    .update(expiry=F('expiry') - (F('expiry') + timedelta(hours=1)))
            users_expired_token_pks.update({username: expired_tokens_of_user})

        admin = self.create_test_admin()  # This line is create 1 user and 1 extra token
        ac, re = self.login(admin)
        with self.assertNumQueries(3):
            """
             Expected queries:
             1- Retrieve user by ID from 'user_id' claim
             2- Retrieve knox token by 'jtk' claim
             3- Get average of counts of created tokens for every user
            """
            response = self.with_token(ac).client.get(average_expired_per_user_url)

        # Calculating average
        expired_token_len_per_user = [len(pks) for name, pks in users_expired_token_pks.items()]
        average = sum(expired_token_len_per_user) / users_count

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['Average expired'], average)

    def test_average_active_per_user(self):
        """ Test that 'average active per user' endpoint is getting average count of active tokens per user"""
        users_list = []  # ['username1', 'username2', .. ]
        users_tokens = {}  # {'username' :[id1, id2, id3, ...]},{},{}
        expiry = timedelta(hours=10)
        users_count = random.randint(1, 20)
        for i in range(0, users_count):  # Create user
            user = self.create_test_user()
            users_list.append(user.username)
            pks = []
            for j in range(0, random.randint(1, 50)):
                instance, _ = AuthToken.objects.create(user, expiry)
                pks.append(instance.pk)

            users_tokens.update({user.username: pks})

        for username in users_list:  # Expire random token of users
            active_tokens_of_user = users_tokens[username]
            random_index = random.sample(range(0, len(active_tokens_of_user)), random.randint(1, len(active_tokens_of_user)))
            for index in sorted(random_index, reverse=True):
                chosen_pk_of_user = active_tokens_of_user.pop(index)
                # Expire token
                AuthToken.objects \
                    .filter(pk=chosen_pk_of_user) \
                    .update(expiry=F('expiry') - (F('expiry') + timedelta(hours=1)))
            if len(active_tokens_of_user) == 0:
                del users_tokens[username]
                users_list.remove(username)

        admin = self.create_test_admin()  # This line is create 1 user and 1 extra token
        ac, re = self.login(admin)
        with self.assertNumQueries(3):
            """
             Expected queries:
             1- Retrieve user by ID from 'user_id' claim
             2- Retrieve knox token by 'jtk' claim
             3- Get average of active token counts for every user
            """
            response = self.with_token(ac).client.get(average_active_per_user_url)

        # Calculating average
        active_token_len_per_user = [len(pks) for name, pks in users_tokens.items()]
        average = sum(active_token_len_per_user) / len(users_list)

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.data['Average active'], average)



from django.utils.translation import gettext_lazy as _
import jwt
from jwt import InvalidAlgorithmError, InvalidTokenError, PyJWKClient, algorithms

from .exceptions import TokenBackendError
from .utils import format_lazy

ALLOWED_ALGORITHMS = (
    'HS256',
    'HS384',
    'HS512',
    'RS256',
    'RS384',
    'RS512',
)


class TokenBackend:
    def __init__(
        self,
        algorithm,
        signing_key=None,
        verifying_key=None,
        audience=None,
        issuer=None,
        jwk_url: str = None,
        leeway=0,
    ):
        self._validate_algorithm(algorithm)

        self.algorithm = algorithm
        self.signing_key = signing_key
        self.audience = audience
        self.issuer = issuer

        self.jwks_client = PyJWKClient(jwk_url) if jwk_url else None
        self.leeway = leeway

        if algorithm.startswith("HS"):
            self.verifying_key = signing_key
        else:
            self.verifying_key = verifying_key

    def _validate_algorithm(self, algorithm):
        """
        Ensure that the nominated algorithm is recognized, and that cryptography is installed for those
        algorithms that require it
        """
        if algorithm not in ALLOWED_ALGORITHMS:
            raise TokenBackendError(format_lazy(_("Unrecognized algorithm type '{}'"), algorithm))

        if algorithm in algorithms.requires_cryptography and not algorithms.has_crypto:
            raise TokenBackendError(format_lazy(_("You must have cryptography installed to use {}."), algorithm))

    def get_verifying_key(self, token):
        if self.algorithm.startswith("HS"):
            return self.signing_key

        if self.jwks_client:
            return self.jwks_client.get_signing_key_from_jwt(token).key

        return self.verifying_key

    def encode(self, payload):
        """
        Returns an encoded token for the given payload dictionary.
        """
        jwt_payload = payload.copy()
        if self.audience is not None:
            jwt_payload['aud'] = self.audience
        if self.issuer is not None:
            jwt_payload['iss'] = self.issuer

        token = jwt.encode(jwt_payload, self.signing_key, algorithm=self.algorithm)
        if isinstance(token, bytes):
            # For PyJWT <= 1.7.1
            return token.decode('utf-8')
        # For PyJWT >= 2.0.0a1
        return token

    def decode(self, token, verify=True):
        """
        Performs a validation of the given token and returns its payload
        dictionary.

        Raises a `TokenBackendError` if the token is malformed, if its
        signature check fails, or if its 'exp' claim indicates it has expired.
        """
        try:
            return jwt.decode(
                token,
                self.get_verifying_key(token),
                algorithms=[self.algorithm],
                audience=self.audience,
                issuer=self.issuer,
                leeway=self.leeway,
                options={
                    'verify_aud': self.audience is not None,
                    'verify_signature': verify,
                },
            )
        except InvalidAlgorithmError as ex:
            raise TokenBackendError(_('Invalid algorithm specified')) from ex
        except InvalidTokenError:
            raise TokenBackendError(_('Token is invalid or expired'))



from datetime import datetime, timedelta
from uuid import UUID, uuid4
from binascii import Error as binascii_Error

from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from knox.crypto import hash_token
from knox.models import AuthToken
from knox.settings import CONSTANTS

from jw_nx.backends import TokenBackend
from jw_nx.exceptions import TokenBackendError, TokenError
from jw_nx.settings import api_settings
from jw_nx.utils import aware_utc_now, datetime_to_unix, datetime_from_unix

USER = get_user_model()
USER_ID_FIELD = api_settings.JW_NX_USER_ID_FIELD
try:
    from hmac import compare_digest
except ImportError:
    def compare_digest(a, b):
        return a == b


class Token:
    """
    A class which validates and wraps an existing JWT or can be used to build a
    new JWT.
    """
    token_type = NotImplemented
    lifetime = NotImplemented
    claims = NotImplemented

    def __init__(self, current_time=None, *args, **kwargs):
        """
        !!!! IMPORTANT !!!! MUST raise a TokenError with a user-facing error
        message if the given token is invalid, expired, or otherwise not safe
        to use.
        """
        self.current_time = current_time if current_time is not None else aware_utc_now()
        self.token_backend = self.get_token_backend()
        self.payload = {}

    def __repr__(self):
        return repr(self.payload)

    def __getitem__(self, key):
        return self.payload[key]

    def __setitem__(self, key, value):
        self.payload[key] = value

    def __delitem__(self, key):
        del self.payload[key]

    def __contains__(self, key):
        return key in self.payload

    def get(self, key, default=None):
        return self.payload.get(key, default)

    @staticmethod
    def get_token_backend():
        return TokenBackend(api_settings.JW_NX_ALGORITHM, api_settings.JW_NX_SIGNING_KEY,
                            api_settings.JW_NX_VERIFYING_KEY, api_settings.JW_NX_AUDIENCE,
                            api_settings.JW_NX_ISSUER, api_settings.JW_NX_JWK_URL,
                            api_settings.JW_NX_LEEWAY)

    def __str__(self):
        """ Signs and returns a token as a base64 encoded string. """
        return self.token_backend.encode(self.payload)

    def verify(self):
        """
        Performs additional validation steps which were not performed when this
        token was decoded.  This method is part of the "public" API to indicate
        the intention that it may be overridden in subclasses.
        """
        self.verify_exp()

        # Ensure token id is present
        if 'jti' not in self.payload:
            raise TokenError(_('Token has no id'))

        self.verify_token_type()

    def verify_payload(self):
        for claim in self.payload:
            if claim not in self.claims:
                msg = _(f"Invalid claims in payload: {claim}")
                raise TokenError(msg)

    def verify_token_type(self):
        """
        Ensures that the token type claim is present and has the correct value.
        """
        try:
            token_type = self.payload['token_type']
        except KeyError:
            raise TokenError(_('Token has no type'))

        if self.token_type != token_type:
            raise TokenError(_('Token has wrong type'))

    def verify_exp(self):
        """
        Verify whether a timestamp value in the given claim has passed (since
        the given datetime value in `self.current_time`).
        Raises a TokenError with a user-facing error message if so.
        """
        claim = 'exp'
        try:
            claim_value = self.payload[claim]
        except KeyError:
            msg = _(f"Token has no '{claim}' claim")
            raise TokenError(msg)

        claim_time = datetime_from_unix(claim_value)
        if claim_time <= self.current_time:
            msg = _(f"Token '{claim}' claim has expired")
            raise TokenError(msg)

    def verify_user_id_field(self):
        user_id = self.payload['user_id']

        if not isinstance(user_id, int) or user_id < 1:
            msg = _(f"Invalid 'user_id' claim")
            raise TokenError(msg)

    def verify_jkt(self):
        knox_token = self.get('jkt')
        if isinstance(knox_token, str): knox_token = bytes(knox_token, 'utf-8')
        matched_token = None
        token = knox_token.decode()
        auth_tokens = AuthToken.objects \
            .filter(token_key=token[:CONSTANTS.TOKEN_KEY_LENGTH],
                    expiry__gte=aware_utc_now(), user_id=self.get('user_id'))
        for auth_token in auth_tokens:
            try:
                digest = hash_token(token, auth_token.salt)
            except (TypeError, binascii_Error):
                msg = _('Error in getting hash token')
                raise TokenError(msg)
            if digest == auth_token.digest:
                matched_token = auth_token
                break

        if matched_token is None:
            msg = _("Invalid token")
            raise TokenError(msg)
        return matched_token

    def set_token_type(self, claim='token_type'):
        self.payload[claim] = self.token_type

    def set_exp(self, claim='exp'):
        """
        Updates the expiration time of a token.

        See here:
        https://tools.ietf.org/html/rfc7519#section-4.1.4
        """
        lifetime: timedelta = self.lifetime
        from_time: datetime = self.current_time
        self.payload[claim] = datetime_to_unix(from_time + lifetime)

    def set_iat(self, claim='iat'):
        """
        Updates the time at which the token was issued.

        See here:
        https://tools.ietf.org/html/rfc7519#section-4.1.6
        """
        self.payload[claim] = datetime_to_unix(self.current_time)

    def set_jkt(self, user, jkt=None, claim='jkt'):
        """ Set JwtKnoxToken=jkt to payload """
        if jkt is None:
            _, jkt = AuthToken.objects.create(user, expiry=self.lifetime)
            bytes(jkt, 'utf-8')

        self.payload[claim] = jkt

    def set_user_id_field(self, user_id):
        """
        Set user_id to payload
        """
        self.payload[f'user_{USER_ID_FIELD}'] = user_id

    def check_token_life(self):
        if (self.token_type or self.lifetime) is not NotImplemented:
            raise TokenError(_('Cannot create token with no type or lifetime'))

    def decode(self, token):
        try:
            return self.token_backend.decode(token=token, verify=True)
        except TokenBackendError:
            raise TokenError(_('Token is invalid or expired'))


class AccessToken(Token):
    token_type = 'access'
    lifetime = api_settings.JW_NX_ACCESS_TOKEN_LIFETIME
    claims = ('token_type', 'exp', f'user_id', 'jti', 'jkt')

    def set_jti(self):
        """
        Populates the configured jti claim of a token with a string where there
        is a negligible probability that the same string will be chosen at a
        later time.

        See here:
        https://tools.ietf.org/html/rfc7519#section-4.1.7
        """
        self.payload['jti'] = uuid4().hex

    def verify_jti(self):
        try:
            UUID(self.payload['jti'], version=4)
        except KeyError:
            msg = _(f"Token has no 'jti' claim")
            raise TokenError(msg)
        except ValueError:
            msg = _("Invalid 'jti' claim")
            raise TokenError(msg)

    def verify_user(self, user):
        if self.payload['user_id'] != user.id:
            msg = _("User with user_id claim is not match together")
            raise TokenError(msg)

    def create_token(self, user, jkt):
        self.set_token_type()
        self.set_exp()
        self.set_jti()
        self.set_user_id_field(user.id)
        self.set_jkt(user, jkt)

    def validate_token(self, token, user):
        self.payload = self.decode(token)
        self.verify_payload()
        self.verify_token_type()
        self.verify_user_id_field()
        self.verify_user(user)
        self.verify_exp()
        self.verify_jti()
        self.verify_jkt()


class RefreshToken(Token):
    token_type = 'refresh'
    lifetime = api_settings.JW_NX_REFRESH_TOKEN_LIFETIME
    claims = ('token_type', 'exp', 'iat', 'jkt', f'user_id')
    is_verified: bool = False

    def verify_iat(self):
        try:
            claim_value = self.payload['iat']
        except KeyError:
            msg = _(f"Token has no 'iat' claim")
            raise TokenError(msg)
        if not isinstance(claim_value, int):
            msg = _("Invalid 'iat' claim")
            raise TokenError(msg)

    def create_token(self, user):
        self.user = user
        self.set_token_type()
        self.set_exp()
        self.set_iat()
        self.set_user_id_field(user.id)
        self.set_jkt(user)
        self.is_verified = True

    def validate_token(self, token):
        self.payload = self.decode(token)
        self.verify_token_type()
        self.verify_user_id_field()
        self.verify_exp()
        self.verify_iat()
        self.verify_jkt()
        self.is_verified = True

    @property
    def access_token(self):
        if not self.is_verified:
            raise TokenBackendError("Refresh token is not verified.")
        if not hasattr(self, 'user'):
            self.user = USER.objects.filter(id=self.payload['user_id']).only('id').first()
        access = AccessToken(current_time=self.current_time)
        access.create_token(self.user, self.payload['jkt'])
        return access

    def create_r_a(self, user):
        """ Create refresh token and access token at the same time """
        self.create_token(user)
        access = self.access_token
        return self, access



from typing import Optional

from django.conf import settings
from django.utils.functional import lazy
from django.utils.module_loading import import_string
from django.utils.timezone import is_naive, make_aware, utc
from jwt import encode as jwt_encode, decode as jwt_decode
from uuid import UUID
from calendar import timegm
from datetime import datetime, timedelta

from django.contrib.auth import get_user_model
from knox.models import AuthToken, User

from jw_nx.settings import api_settings

sha = import_string('cryptography.hazmat.primitives.hashes.SHA512')


def make_utc(dt):
    if settings.USE_TZ and is_naive(dt):
        return make_aware(dt, timezone=utc)

    return dt


def aware_utc_now():
    return make_utc(datetime.utcnow())


def datetime_to_unix(dt):
    return timegm(dt.utctimetuple())


def datetime_from_unix(ts):
    return make_utc(datetime.utcfromtimestamp(ts))


def jwt_encode_handler(payload):
    return jwt_encode(payload, api_settings.JW_NX_SECRET_KEY, api_settings.JW_NX_ALGORITHM)


def jwt_join_header_and_token(token):
    return f"Bearer {token}"


def jwt_response_payload_handler(token):
    return {'token': jwt_join_header_and_token(token), }


def get_username_field():
    try:
        return get_user_model().USERNAME_FIELD
    finally:
        return 'username'


def get_username(user):
    try:
        return user.get_username()
    finally:
        return user.username


def create_auth_token(user, expiry: timedelta):
    _, token = AuthToken.objects.create(user=user, expiry=expiry)  # Create knox token
    payload = jwt_payload_handler(user, token, expiry)

    return jwt_encode_handler(payload)


def jwt_decode_handler(token):
    options = {'verify_exp': True, }

    return jwt_decode(
        token,
        api_settings.JW_NX_SECRET_KEY,
        algorithms=api_settings.JW_NX_ALGORITHM,
        options=options,
        leeway=api_settings.JW_NX_LEEWAY,
        audience=api_settings.JW_NX_AUDIENCE,
        issuer=api_settings.JW_NX_ISSUER,
    )


def jwt_payload_handler(user: User, token: str, expiry: Optional[timedelta]):
    username_field = get_username_field()
    username = get_username(user)
    now = datetime.utcnow()
    payload = {
        username_field: username,
        'iat': timegm(now.utctimetuple()),  # Issued At Time => unix timestamp
        'jti': token,
        # OPTIONALS
        # More info => https://docs.microsoft.com/en-us/azure/active-directory/develop/id-tokens#payload-claims
        # user_id : user.pk
        # exp : now + expiry
        # aud : JWT_AUDIENCE
        # iss : JWT_ISSUER
    }

    if expiry:
        payload['exp'] = now + expiry

    if isinstance(user.pk, UUID):
        payload['user_id'] = str(user.pk)

    jwt_audience = api_settings.JW_NX_AUDIENCE
    if jwt_audience is not None:
        payload['aud'] = jwt_audience

    jwt_issuer = api_settings.JW_NX_ISSUER
    if jwt_issuer is not None:
        payload['iss'] = jwt_issuer

    return payload


def format_lazy(s, *args, **kwargs):
    return s.format(*args, **kwargs)


format_lazy = lazy(format_lazy, str)
